import React from "react";

export function Loader() {
  return (
    <div class="slider">
      <div class="line"></div>
      <div class="subline inc"></div>
      <div class="subline dec"></div>
    </div>
  );
}

export default Loader;
