<html>
<head>
    <title>Login</title>
    <script>
        {% include 'js/common.js' %}
    </script>
    <style>
        {% include 'css/common.css' %}
    </style>
</head>
    <body>
        <div class="centered">
            {% if status_code is defined %}
            <div class="error">
                <span class="error-code">Error {{ status_code }}</span>
                <span class="error-message">{{ detail }}</span>
            </div>
            {% endif %}
            <h1>Login</h1>
            <form class="centered-form" method="get" action="/api/v1/auth/login">
                <input type="password" name="{{ api_key_name }}" id="{{ api_key_name }}" placeholder="API Key...">
                <input type="submit" value="Login">
            </form>
        </div>
    </body>
</html>